% release version of towersOfGold
% 
% Associates gold particles to towers, registers towers/gold part. across rows,
% normalizes by average height across rows, saves convex hull of normalized towers
% and gold particles on independent CSV tables on directory of input files.
%
% inputs:
%   GoldParticles.am
%   LinkTable.csv
%   Towers.am
%   TowerUsage.csv


%%

clear, clc
tic

%% setup parameters

dataPath = '~/Downloads/ToG_Data/Cells/50687';


resizeFactor = 1/4; % volume will be resized by this amount to speed-up computations
borderSamplingFactor = 1/10; % not all border voxels necessary for some computations and visualizations
gold2towerAssociationMethod = 'NN'; % options: 'NN' (nearest neighbors, fast), 'DT' (distance transform, slow)
nNearestNeighbors = 10; % for gold2tower association when 'NN' option is chosen
minDistThr = 100; % in microns; particles such that distance to nearest tower > this threshold are excluded

%% setup output folders

towersFilePath = [dataPath filesep 'Towers.am'];
goldPartFilePath = [dataPath filesep 'GoldParticles.am'];
tLinkTablePath = [dataPath filesep 'LinkTable.csv'];
usageTablePath = [dataPath filesep 'TowerUsage.csv'];

outFiguresPath = [dataPath filesep 'Figures'];
outTablesPath = [dataPath filesep 'Tables'];
if ~(exist(outFiguresPath,'dir') == 7)
    mkdir(outFiguresPath);
end
if ~(exist(outTablesPath,'dir') == 7)
    mkdir(outTablesPath);
end

%% read stack
disp('read stack')

ImageData = bfopen(towersFilePath);

%% get metadata

import java.util.Hashtable
h = ImageData{2};
xVoxPerMet = h.get('Global Pixels per meter (X)');
yVoxPerMet = h.get('Global Pixels per meter (Y)');
zVoxPerMet = h.get('Global Pixels per meter (Z)');
zStretch = xVoxPerMet/zVoxPerMet;
xVoxSize = 1/xVoxPerMet; % in meters
yVoxSize = 1/yVoxPerMet;
zVoxSize = 1/zVoxPerMet;
fprintf('xVoxelSize: %f\nyVoxelSize: %f\nzVoxelSize: %f\n(meters)\n', xVoxSize, yVoxSize, zVoxSize)

%% grab planes
disp('grab planes')

nPlanes = size(ImageData{1},1);
[nRows,nCols] = size(ImageData{1}{1,1});

V = uint16(zeros(nRows,nCols,nPlanes));

for k = 1:nPlanes
    fprintf('.')
    if mod(k,75) == 0
        fprintf('\n')
    end
    V(:,:,k) = ImageData{1}{k,1};
end
fprintf('\n')
clear ImageData
fprintf('found %d planes of %dx%d pixels\n', size(V,3), size(V,1),size(V,2))

%% resize, remove interior
disp('resize, remove interior')

RV = imresize3(V,resizeFactor,'nearest');
EV = imerode(RV > 0,strel('sphere',1));
HRV = RV;
HRV(EV) = 0;
clear V EV

%% read stack (gold particles)
disp('read stack (gold particles)')

ImageDataGP = bfopen(goldPartFilePath);

%% grab planes (gold particles)
disp('grab planes (gold particles)')

nPlanes = size(ImageDataGP{1},1);
[nRows,nCols] = size(ImageDataGP{1}{1,1});

G = false(nRows,nCols,nPlanes);

for k = 1:nPlanes
    fprintf('.')
    if mod(k,75) == 0
        fprintf('\n')
    end
    G(:,:,k) = ImageDataGP{1}{k,1} > 0;
end
fprintf('\n')
clear ImageDataGP

%% keep only centroids of gold particle blobs
disp('keep only centroids of gold particle blobs')

stats = regionprops(G,'Centroid');
clear G
centroids = resizeFactor*cat(1,stats.Centroid);
% centroid coordinates: j, i, k (i.e., x,y,z, where x is column, y is row)

%% sample border (we dont need all voxels)
disp('sample border')

S = double(HRV).*(rand(size(HRV)) < borderSamplingFactor);

%% visualize point cloud with gold particles

[i,j,k] = ind2sub(size(S),find(S));
jg = centroids(:,1); ig = centroids(:,2); kg = centroids(:,3);
k = k*zStretch;
kg = kg*zStretch;

scsz = get(0,'ScreenSize'); % scsz = [left botton width height]
h = figure('Position',[100 100 1000 500]);

plot3(j,i,k,'.','Color',0.5*[1 1 1],'MarkerSize',1), hold on
plot3(jg,ig,kg,'.','Color',[1 1 0]), hold off
xlabel('x')
ylabel('y')
zlabel('z')
axis equal, axis off
title('sampled tower points, and gold particles')
% rotate3d on


h.InvertHardcopy = 'off';
h.Color = 'black';

view(185,-85)

saveas(gcf,[outFiguresPath filesep sprintf('1_Data.fig')]);
close all

%% get tower point clouds
disp('get tower point clouds')

nObjects = max(HRV(:));
TowerPointClouds = cell(1,nObjects);
for index = 1:nObjects
    fprintf('.')
    if mod(index,75) == 0
        fprintf('\n')
    end
    VS = S.*(HRV == index);
    [i,j,k] = ind2sub(size(VS),find(VS));
    k = k*zStretch;
    TowerPointClouds{index} = [j i k]; % x, y, z (need this re-ordering for plot)
end
fprintf('\n')

%% find out towers principal directions
disp('find out towers principal directions')

% individual directions
dirs = [];
% figure, hold on
for index = 1:nObjects
    fprintf('.')
    if mod(index,75) == 0
        fprintf('\n')
    end
    TPC = TowerPointClouds{index};
    if ~isempty(TPC)
        j = TPC(:,1); i = TPC(:,2); k = TPC(:,3);
        mi = mean(i);
        mj = mean(j);
        mk = mean(k);
        i = i-mi;
        j = j-mj; 
        k = k-mk;
        
        coeff = pca([j i k]);
        flipCoeff = (-1)^(rand < 0.5);
        coeff(:,1) = flipCoeff*coeff(:,1);
        dirs = [dirs coeff(:,1)];
        
%         plot3(mi+i,mj+j,mk+k,'.k','MarkerSize',0.5)%, hold on
%         plot3(mi+200*[0 coeff(2,1)],mj+200*[0 coeff(1,1)],mk+200*[0 coeff(3,1)],'-r','MarkerSize',5)%, hold off
%         axis equal
%         xlabel('i'), ylabel('j'), zlabel('k')
%         pause(0.1)
    end
end
% fprintf('\n')
close all

%% find direction cluster centers (via mean shift on sphere)

kms = 2;
% kms can be thought of as inverse of 'bandwidth' in mean shift literature parlance
% rule of thumb: for n equally spaced clusters on the 2D disk set k = n
[c,l] = vmf(dirs',kms);

scsz = get(0,'ScreenSize'); % scsz = [left botton width height]
h = figure('Position',[100 100 500 500]);

hold on
[x,y,z] = sphere;
surf(x,y,z), alpha(0.25)
for index = 1:size(c,1)
    h = (index-1)/size(c,1)*2/3;
    plot3(dirs(1,l == index),dirs(2,l == index),dirs(3,l == index),'.','Color',hsv2rgb([h 1 1]),'MarkerSize',10)
    plot3(c(index,1),c(index,2),c(index,3),'s','Color',hsv2rgb([h 1 1]),'MarkerSize',10)
    plot3(c(index,1),c(index,2),c(index,3),'.k','MarkerSize',10)
end
hold off
axis equal, axis([-1 1 -1 1 -1 1])
% title(sprintf('%d directions found', size(c,1)))
% rotate3d on
axis off

view(c(1,:))

saveas(gcf,[outFiguresPath filesep sprintf('2_PrincipalDirections.fig')]);
close all

%% decide which direction is *up*
disp('decide which direction is *up*')

[ii,jj,kk] = ind2sub(size(S),find(S));
kk = kk*zStretch;
m = [mean(ii), mean(jj), mean(kk)];
ds = nan(nObjects,size(c,1));
for index = 1:nObjects
    fprintf('.')
    if mod(index,75) == 0
        fprintf('\n')
    end
    TPC = TowerPointClouds{index};
    if ~isempty(TPC)
        j = TPC(:,1); i = TPC(:,2); k = TPC(:,3);
        i = i-m(1); j = j-m(2); k = k-m(3);
        for iDirection = 1:size(c,1)
            v = c(iDirection,:);
            d = min(sum([j i k].*repmat(v,[length(i) 1]),2));
            ds(index,iDirection) = d;
        end
    end
end
fprintf('\n')

[~,iDirection] = min(std(ds,'omitnan'));
v = c(iDirection,:);
ii = ii-m(1); jj = jj-m(2); kk = kk-m(3);
d = max(sum([jj ii kk].*repmat(v,[length(ii) 1]),2));

scsz = get(0,'ScreenSize'); % scsz = [left botton width height]
h = figure('Position',[100 100 1000 500]);

plot3(jj,ii,kk,'.','Color',0.5*[1 1 1],'MarkerSize',1), hold on
plot3(0,0,0,'oy')
plot3([0 d*v(1)],[0 d*v(2)],[0 d*v(3)],'-y'), hold off
axis equal, axis off
% view(v)
view(-180,-90)
title('view from above')
rotate3d on

h.InvertHardcopy = 'off';
h.Color = 'black';

saveas(gcf,[outFiguresPath filesep sprintf('3_UpDirection.fig')]);
close all

%% associate gold particles to towers
disp('associate gold particles to towers')

if strcmp(gold2towerAssociationMethod,'NN')
    [ii,jj,kk] = ind2sub(size(S),find(S));
    vv = zeros(1,length(ii));
    for index = 1:length(ii)
        vv(index) = HRV(ii(index),jj(index),kk(index));
    end
    kk = kk*zStretch;
    jg = centroids(:,1); ig = centroids(:,2); kg = centroids(:,3);
    kg = kg*zStretch;

    [idx,dst] = knnsearch([jj ii kk],[jg ig kg],'K',nNearestNeighbors);

    SplitGoldParticles = cell(1,nObjects);
    SplitGoldParticlesMinDists = cell(1,nObjects);

    towerIndicesNN = zeros(length(ig),1);
    minDistsNN = zeros(length(ig),1);
    for i = 1:length(ig)
        labels = zeros(1,nNearestNeighbors);
        for j = 1:nNearestNeighbors
            labels(j) = vv(idx(i,j));
        end
        ml = mode(labels);
        towerIndicesNN(i) = ml;
        minDistsNN(i) = min(dst(i,labels == ml));
%         fprintf('%d: %d\n',i,sum(labels == ml));
        gps = SplitGoldParticles{ml};
        gpsd = SplitGoldParticlesMinDists{ml};
        gps = [gps; [jg(i) ig(i) kg(i)]];
        gpsd = [gpsd; minDistsNN(i)];
        SplitGoldParticles{ml} = gps;
        SplitGoldParticlesMinDists{ml} = gpsd;
    end
    minDists = minDistsNN;
elseif strcmp(gold2towerAssociationMethod,'DT')
    nParticles = size(centroids,1);
    distPartObj = inf(nParticles,nObjects);
    RRV = imresize3(RV,[size(RV,1),size(RV,2),zStretch*size(RV,3)],'nearest');
    [ni,nj,nk] = size(RRV);
    jg = centroids(:,1); ig = centroids(:,2); kg = centroids(:,3);
    kg = kg*zStretch;
    for j = 1:nObjects
        fprintf('.')
        if mod(j,75) == 0
            fprintf('\n')
        end
        if ~isempty(TowerPointClouds{j})
            D = bwdist(RRV == j);
            for i = 1:nParticles
                ijk = round([ig(i) jg(i) kg(i)]);
                ijk(1) = min(max(ijk(1),1),ni);
                ijk(2) = min(max(ijk(2),1),nj);
                ijk(3) = min(max(ijk(3),1),nk);
                distPartObj(i,j) = D(ijk(1),ijk(2),ijk(3));
            end
        end
    end
    fprintf('\n')

    towerIndicesDT = zeros(length(ig),1);
    minDistsDT = zeros(length(ig),1);
    SplitGoldParticles = cell(1,nObjects);
    SplitGoldParticlesMinDists = cell(1,nObjects);
    for i = 1:nParticles
        [minDistsDT(i),im] = min(distPartObj(i,:));
        gps = SplitGoldParticles{im};
        gpsd = SplitGoldParticlesMinDists{im};
        gps = [gps; [jg(i) ig(i) kg(i)]];
        gpsd = [gpsd; minDistsDT(i)];
        SplitGoldParticles{im} = gps;
        SplitGoldParticlesMinDists{im} = gpsd;
        towerIndicesDT(i) = im;
    end
    minDists = minDistsDT;
end

%% display towers and associated gold particles
disp('display towers and associated gold particles')

scsz = get(0,'ScreenSize'); % scsz = [left botton width height]
h = figure('Position',[100 100 1000 500]);

hold on
hues = double(0:nObjects-1)/double(nObjects)*2/3;
hues = hues(randperm(length(hues)));
for index = 1:nObjects
    TPC = TowerPointClouds{index};
    if ~isempty(TPC)
        plot3(TPC(:,1),TPC(:,2),TPC(:,3),'.','Color',hsv2rgb([hues(index) 1 0.5]),'MarkerSize',1)
        SGP = SplitGoldParticles{index};
        if ~isempty(SGP)
            plot3(SGP(:,1),SGP(:,2),SGP(:,3),'*','Color',hsv2rgb([hues(index) 1 1]))
        end
    end
end
hold off
axis equal
axis off
view(180,-90)
title('towers and associated gold particles')
% rotate3d on

h.InvertHardcopy = 'off';
h.Color = 'black';

saveas(gcf,[outFiguresPath filesep sprintf('4_TowersAndAssociatedGoldPart.fig')]);
close all

%% align towers, gold particles to *up* direction
disp('align towers, gold particles to *up* direction')

w = c(iDirection,:)';
Tjik = [];
T = cell(1,nObjects);
GP = cell(1,nObjects);
for index = 1:nObjects
    fprintf('.')
    if mod(index,75) == 0
        fprintf('\n')
    end
    TPC = TowerPointClouds{index};
    if ~isempty(TPC)
        j = TPC(:,1); i = TPC(:,2); k = TPC(:,3);
        mi = mean(i); mj = mean(j); mk = mean(k);
        i = i-mi;     j = j-mj;     k = k-mk;
        
        evecs = pca([j i k]);

        if dot(w,evecs(:,1)) < 0 % pointing to wrong direction
            v = -evecs(:,1);
        else
            v = evecs(:,1);
        end
        
        v1 = v; % positive direction of principal component
        v2 = w; % 'up' direction
        
        [unitk,theta,v1Rotated] = vregister(v1,v2);

        tjik = zeros(length(i),3);
        for l = 1:length(i)
           vRotated = axangrot([j(l); i(l); k(l)],unitk,theta);
           tjik(l,:) = vRotated';
        end
        tjik(:,1) = tjik(:,1)+mj;
        tjik(:,2) = tjik(:,2)+mi;
        tjik(:,3) = tjik(:,3)+mk;
        Tjik = [Tjik; tjik];
        T{index} = tjik;
        
        SGP = SplitGoldParticles{index};
        if ~isempty(SGP)
            j = SGP(:,1);
            i = SGP(:,2);
            k = SGP(:,3);
            i = i-mi;
            j = j-mj;
            k = k-mk;
            tjik = zeros(length(i),3);
            for l = 1:length(i)
               vRotated = axangrot([j(l); i(l); k(l)],unitk,theta);
               tjik(l,:) = vRotated';
            end
            tjik(:,1) = tjik(:,1)+mj;
            tjik(:,2) = tjik(:,2)+mi;
            tjik(:,3) = tjik(:,3)+mk;
            GP{index} = tjik;
        end
    end
end
fprintf('\n')

figureBB(1000,500)

hold on
for index = 1:nObjects
    t = T{index};
    if ~isempty(t)
        plot3(t(:,1),t(:,2),t(:,3),'.','Color',hsv2rgb([hues(index) 1 0.5]),'MarkerSize',1)
    end
    t = GP{index};
    if ~isempty(t)
        plot3(t(:,1),t(:,2),t(:,3),'*','Color',hsv2rgb([hues(index) 1 1]))
    end
end
hold off
axis equal
% axis([min(t(:,1)) max(t(:,1)) min(t(:,2)) max(t(:,2)) min(t(:,3)) max(t(:,3))])
view(180,-90)
axis off
title('towers aligned upwards')


saveas(gcf,[outFiguresPath filesep sprintf('5_UpDirectionAlignment.fig')]);
close all


%% make *up* direction [0 0 1], floor z = 0, measure height
disp('make *up* direction [0 0 1], floor z = 0, measure height')

v1 = w; % 'up' direction
v2 = [0; 0; 1];

[unitk,theta,v1Rotated] = vregister(v1,v2);

meanTjik = mean(Tjik);

NT = cell(1,nObjects);
NGP = cell(1,nObjects);
mts = cell(1,nObjects);
tHeights = cell(1,nObjects);
tRadii = cell(1,nObjects);

figureBB(1500,750)

hold on
for index = 1:nObjects
    fprintf('.')
    if mod(index,75) == 0
        fprintf('\n')
    end
    t = T{index};
    if ~isempty(t)
        t = t-repmat(meanTjik,[size(t,1) 1]);
        for iv = 1:size(t,1)
            vRotated = axangrot(t(iv,:)',unitk,theta);
            t(iv,:) = vRotated';
        end
        tMin = min(t(:,3));
        t(:,3) = t(:,3)-tMin;
        tHeights{index} = max(t(:,3));
        NT{index} = t;
        plot3(t(:,1),t(:,2),t(:,3),'.','Color',hsv2rgb([hues(index) 1 0.5]))
        mt = mean(t);
        mt(3) = 0;
        mts{index} = mt';
        text(mt(1),mt(2),mt(3)+300,sprintf('%d',index),'Color','white');
        tt = t(:,1:2)-repmat(mt(1:2),[size(t,1) 1]);
        tRadii{index} = median(sqrt(tt(:,1).^2+tt(:,2).^2));
    end
    
    t = GP{index};
    if ~isempty(t)
        t = t-repmat(meanTjik,[size(t,1) 1]);
        for iv = 1:size(t,1)
            vRotated = axangrot(t(iv,:)',unitk,theta);
            t(iv,:) = vRotated';
        end
        t(:,3) = t(:,3)-tMin;
        NGP{index} = t;
        plot3(t(:,1),t(:,2),t(:,3),'*','Color',hsv2rgb([hues(index) 1 1]))
    end
end
fprintf('\n')

% allT = [];
% for i = 1:length(NGP)
%     allT = [allT; NGP{i}];
% end


hold off
axis equal
% view([0 0 1])
% axis([0.5*min(allT(:,1)) 0.5*max(allT(:,1)) min(allT(:,2)) max(allT(:,2)) min(allT(:,3)) max(allT(:,3))])
view(-170,90)
xlabel('x'), ylabel('y'), zlabel('z')
axis off
% title('floor z = 0, up [0 0 1]')
saveas(gcf,[outFiguresPath filesep sprintf('6_Up001FloorZeq0.fig')]);
close all


%% [optional] visualize individual upright towers with associated particles

if 0
figureBB(1000,500)
for index = 32%1:nObjects
    clf
    t = NT{index};
    if ~isempty(t)
        I = eye(3);
        for j = 1:3
            subplot(1,3,j)
            t = NT{index};
            plot3(t(:,1),t(:,2),t(:,3),'.','Color','white','MarkerSize',4)
            hold on
            t = NGP{index};
            if ~isempty(t)
                plot3(t(:,1),t(:,2),t(:,3),'*','Color','yellow')
            end
            hold off
            axis equal
            view(I(j,:))
            set(gca,'Color','black')
%             xlabel('x'), ylabel('y'), zlabel('z')
            axis off
            if j == 3
                title(sprintf('%d',index))
            end
        end
%         pause
    end
end
saveas(gcf,[outFiguresPath filesep sprintf('6B_Up001FloorZeq0_T32.fig')]);
close all
end

%% compute azimuth directions for individual towers
disp('compute azimuth directions for individual towers')

tLinkTable = readtable(tLinkTablePath);
tlt = table2array(tLinkTable); % tower link table
usageTable = readtable(usageTablePath);
ut = table2array(usageTable);

uniqueDirectionBefore = cell(1,nObjects);
rowIDs = cell(1,nObjects);
useForAverage = zeros(1,nObjects);
for index = 1:nObjects
    col = [];
    for row = 1:4
        col = find(tlt(row,:) == index);
        if ~isempty(col)
            break
        end
    end
    if ~isempty(col)
        useForAverage(index) = ut(row,col);
        if row == 1
            rowIDs{index} = 1;
            mt0 = mts{index};
            if ~isempty(mt0)
                iRow2 = tlt(2,col);
                if iRow2 ~= 0
                    mt1 = mts{iRow2};
                    if ~isempty(mt1)
                        udb = mt1-mt0;
                        udb = udb/norm(udb);
                        uniqueDirectionBefore{index} = udb;
                    end
                end
            end
        elseif row == 2
            rowIDs{index} = 2;
            mt0 = mts{index};
            if ~isempty(mt0)
                iRow3 = tlt(3,col);
                if iRow3 ~= 0
                    mt1 = mts{iRow3};
                    if ~isempty(mt1)
                        udb = mt1-mt0;
                        udb = udb/norm(udb);
                        uniqueDirectionBefore{index} = udb;
                    end
                end
            end
        elseif row == 4
            rowIDs{index} = 4;
            mt1 = mts{index};
            if ~isempty(mt1)
                iRow3 = tlt(3,col);
                if iRow3 ~= 0
                    mt0 = mts{iRow3};
                    if ~isempty(mt0)
                        udb = mt1-mt0;
                        udb = udb/norm(udb);
                        uniqueDirectionBefore{index} = udb;
                    end
                end
            end
        elseif row == 3
            rowIDs{index} = 3;
            mt1A = mts{index}; % row 2 to row 3
            if ~isempty(mt1A)
                iRow2 = tlt(2,col);
                udb23 = [];
                if iRow2 ~= 0
                    mt0A = mts{iRow2};
                    if ~isempty(mt0A)
                        udb23 = mt1A-mt0A;
                    end
                end
            end
            mt0B = mts{index}; % row 3 to row 4
            if ~isempty(mt0B)
                iRow4 = tlt(4,col);
                udb34 = [];
                if iRow4 ~= 0
                    mt1B = mts{iRow4};
                    if ~isempty(mt1B)
                        udb34 = mt1B-mt0B;
                    end
                end
            end
            if ~isempty(udb23) && ~isempty(udb34)
                udb23 = udb23/norm(udb23);
                udb34 = udb34/norm(udb34);
                udb = udb23+udb34;
                uniqueDirectionBefore{index} = udb/norm(udb);
            elseif ~isempty(udb23)
                uniqueDirectionBefore{index} = udb23/norm(udb23);
            elseif ~isempty(udb34)
                uniqueDirectionBefore{index} = udb34/norm(udb34);
            end
        end
    end
end

figureBB(1500,750)

hold on
dd = d/7;
for index = 1:nObjects
    p = mts{index};
    if ~isempty(p)
        plot3(p(1),p(2),p(3),'.r','MarkerSize',16)
        v = uniqueDirectionBefore{index};
        if ~isempty(v)
            plot3([p(1) p(1)+dd*v(1)],[p(2) p(2)+dd*v(2)],[p(3) p(3)+dd*v(3)],'-r')
        end
        text(p(1),p(2),p(3),sprintf('%d',index),'Color','white');
    end
end

hold off
axis equal
% view([0 0 1])
view(-170,90)
xlabel('x'), ylabel('y'), zlabel('z')
title('azimuth directions')
axis off
saveas(gcf,[outFiguresPath filesep sprintf('7_Azimuth.fig')]);
close all


%% compute azimuth transforms for individual towers
disp('compute azimuth transforms for individual towers')

unitkArray = cell(1,nObjects);
thetaArray = cell(1,nObjects);
directionAfter = cell(1,nObjects);


figureBB(1500,750)

hold on
for index = 1:nObjects
    udb = uniqueDirectionBefore{index};
    if ~isempty(udb)
        [unitk,theta,pR] = vregister(udb,[0; 1; 0]);
        directionAfter{index} = pR;
        unitkArray{index} = unitk;
        thetaArray{index} = theta;
        mt = mts{index};
        plot3(mt(1), mt(2), mt(3),'.r','MarkerSize',16)
        plot3([mt(1) mt(1)+dd*udb(1)],[mt(2) mt(2)+dd*udb(2)],[mt(3) mt(3)+dd*udb(3)],'-r','MarkerSize',10)
        plot3([mt(1) mt(1)+dd*pR(1)],[mt(2) mt(2)+dd*pR(2)],[mt(3) mt(3)+dd*pR(3)],'-g','MarkerSize',10)
        text(mt(1),mt(2),mt(3),sprintf('%d',index),'Color','white');
    end
end

hold off
axis equal
% view([0 0 1])
view(-170,90)
xlabel('x'), ylabel('y'), zlabel('z')
title('azimuth transforms')
axis off
saveas(gcf,[outFiguresPath filesep sprintf('8_AzimuthTransform.fig')]);
close all

%% [optional] apply azimuth transforms on individual towers / gold particles

if 0
figureBB(1000,500)
for index = 7%1:nObjects
    v1 = uniqueDirectionBefore{index};
    v2 = directionAfter{index};
    if ~isempty(v2)
        disp(index)
        p = mts{index}';
        t = NT{index};
        t = t-repmat(p,[size(t,1) 1]);
        u = NGP{index};
        if ~isempty(u)
            u = u-repmat(p,[size(u,1) 1]);
        end
        tu = [t; u];
        subplot(1,2,1)
        plot3(t(:,1),t(:,2),t(:,3),'.w'), hold on
        if ~isempty(u)
            plot3(u(:,1),u(:,2),u(:,3),'*y')
        end
        plot3([0 dd*v1(1)],[0 dd*v1(2)],[0 dd*v1(3)],'-r')
        plot3([0 dd*v2(1)],[0 dd*v2(2)],[0 dd*v2(3)],'-g')
        hold off
        axis equal, axis off, axis([min(tu(:,1)) max(tu(:,1)) min(tu(:,2)) max(tu(:,2)) 0 max(tu(:,3))]), view([0 0 1]), xlabel('x'), ylabel('y'), zlabel('z')
        title(sprintf('index: %d | before', index))
        unitk = unitkArray{index};
        theta = thetaArray{index};
        for i = 1:size(t,1)
            t(i,:) = axangrot(t(i,:)',unitk,theta)';
        end
        for i = 1:size(u,1)
            u(i,:) = axangrot(u(i,:)',unitk,theta)';
        end
        tu = [t; u];
        subplot(1,2,2)
        plot3(t(:,1),t(:,2),t(:,3),'.w'), hold on
        if ~isempty(u)
            plot3(u(:,1),u(:,2),u(:,3),'*y')
        end
        plot3([0 0],[0 dd],[0 0],'-r')
        tdd = axangrot([0 dd 0]',unitk,theta);
        plot3([0 tdd(1)],[0 tdd(2)],[0 tdd(3)],'-g'), hold off
        axis equal, axis off, axis([min(tu(:,1)) max(tu(:,1)) min(tu(:,2)) max(tu(:,2)) 0 max(tu(:,3))]), view([0 0 1]), xlabel('x'), ylabel('y'), zlabel('z')
        title(sprintf('index: %d | after', index))
%         pause
    end
end
saveas(gcf,[outFiguresPath filesep sprintf('8B_AzimuthTransform_T7.fig')]);
close all
end

%% apply azimuth transforms on all towers, gold particles
disp('apply azimuth transforms on all towers, gold particles')

ANT = cell(1,nObjects);
ANGP = cell(1,nObjects);

figureBB(1000,500)

hold on
for index = 1:nObjects
    unitk = unitkArray{index};
    if ~isempty(unitk)
        p = mts{index}';
        t = NT{index};
        t = t-repmat(p,[size(t,1) 1]);
        theta = thetaArray{index};
        for i = 1:size(t,1)
            t(i,:) = axangrot(t(i,:)',unitk,theta)';
%             t(i,:) = mapToZCilinder(axangrot(t(i,:)',unitk,theta)',dd);
        end
        t = t+repmat(p,[size(t,1) 1]);
        ANT{index} = t;
        if useForAverage(index)
            plot3(t(:,1),t(:,2),t(:,3),'.','Color',hsv2rgb([hues(index) 1 0.5]))
        else
            plot3(t(:,1),t(:,2),t(:,3),'.','Color',0.5*[1 1 1])
        end
        
        t = NGP{index};
        if ~isempty(t)
            t = t-repmat(p,[size(t,1) 1]);
            for i = 1:size(t,1)
                t(i,:) = axangrot(t(i,:)',unitk,theta)';
%                 t(i,:) = mapToZCilinder(axangrot(t(i,:)',unitk,theta)',dd);
            end
            t = t+repmat(p,[size(t,1) 1]);
            ANGP{index} = t;
            if useForAverage(index)
                plot3(t(:,1),t(:,2),t(:,3),'*','Color',hsv2rgb([hues(index) 1 1]))
            else
                plot3(t(:,1),t(:,2),t(:,3),'*','Color',0.5*[1 1 1])
            end
        end
    end
end

hold off
axis equal
axis off
view([0 0 1])
xlabel('x'), ylabel('y'), zlabel('z')
title('azimuth transformed towers')
saveas(gcf,[outFiguresPath filesep sprintf('9_AzimuthTransformedTowers.fig')]);
close all

%% find out average height/radius per row
disp('find out average height/radius per row')

hr1 = []; rr1 = []; ir1 = [];
hr2 = []; rr2 = []; ir2 = [];
hr3 = []; rr3 = []; ir3 = [];
hr4 = []; rr4 = []; ir4 = [];
for index = 1:nObjects
    if ~isempty(tHeights{index}) && useForAverage(index)
        rID = rowIDs{index};
        if rID == 1
            ir1 = [ir1 index];
            hr1 = [hr1 tHeights{index}];
            rr1 = [rr1 tRadii{index}];
        elseif rID == 2
            ir2 = [ir2 index];
            hr2 = [hr2 tHeights{index}];
            rr2 = [rr2 tRadii{index}];
        elseif rID == 3
            ir3 = [ir3 index];
            hr3 = [hr3 tHeights{index}];
            rr3 = [rr3 tRadii{index}];
        elseif rID == 4
            ir4 = [ir4 index];
            hr4 = [hr4 tHeights{index}];
            rr4 = [rr4 tRadii{index}];
        end
    end
end
ah = [mean(hr1) mean(hr2) mean(hr3) mean(hr4)];
ar = [mean(rr1) mean(rr2) mean(rr3) mean(rr4)];

%% apply azimuth transforms on all towers, gold particles; register to common tower, all and per row
disp('register to common tower, all and per row')

figureBB(1000,500)
axs = [];
rf = 1/resizeFactor*xVoxSize*1000000; % scale to microns
for i = 1:4%5
%     ax = subplot(1,5,i); hold on
    ax = subplot(1,4,i); hold on
    axs = [axs ax];
end
Ts = [];
Gs = [];
ts = [];
AggregatedTowers = cell(4,1);
AggregatedGoldParticles = cell(4,1);
for index = 1:nObjects
    unitk = unitkArray{index};
    if ~isempty(unitk) && useForAverage(index)
        row = rowIDs{index};
        p = mts{index}';
        t = ANT{index};
        
%         subplot(1,5,5)
%         plot3(t(:,1),t(:,2),t(:,3),'.','Color',0.5*[1 1 1],'MarkerSize',5)
        Ts = [Ts; t];
        t = t-repmat(p,[size(t,1) 1]);
        t(:,3) = t(:,3)/tHeights{index}*ah(rowIDs{index});
        t = rf*t;
%         subplot(1,5,row)
        subplot(1,4,row)
        plot3(t(:,1),t(:,2),t(:,3),'.','Color',0.5*[1 1 1],'MarkerSize',1)
        ts = [ts; t];
        AT = AggregatedTowers{row};
        AT = [AT; t];
        AggregatedTowers{row} = AT;
        
        t = ANGP{index};
        if ~isempty(t)
%             subplot(1,5,5)
%             plot3(t(:,1),t(:,2),t(:,3),'*','Color',[1 1 0])
            Gs = [Gs; t];
            t = t-repmat(p,[size(t,1) 1]);
            t(:,3) = t(:,3)/tHeights{index}*ah(rowIDs{index});
            t = rf*t;
            
            % exclude far away gold particles
            gpsd = SplitGoldParticlesMinDists{index}*rf;
            for i = size(t,1):-1:1
                if gpsd(i) > minDistThr
                    t(i,:) = [];
                end
            end
            ts = [ts; t];

%             subplot(1,5,row)
            subplot(1,4,row)
            plot3(t(:,1),t(:,2),t(:,3),'*','Color',[1 1 0])
            ts = [ts; t];
            AGP = AggregatedGoldParticles{row};
            AGP = [AGP; t];
            AggregatedGoldParticles{row} = AGP;
        end
    end
end
mnts = min(ts);
mxts = max(ts);
mmts = max([abs(mnts); abs(mxts)]);
mmtsxy = max(mmts(1:2));

[x,y,z] = cylinder(0.9*mmtsxy*ones(1,5),12);
z = mmts(3)*z/5;

for i = 1:4
%     subplot(1,5,i)
    subplot(1,4,i)
    surf(x,y,z), alpha(0.2)
    plot3([0 0],[0 0.9*mmtsxy],[0 0],'-w','LineWidth',2)
    hold off
    axis equal
    axis([-mmtsxy mmtsxy -mmtsxy mmtsxy 0 mmts(3)])
    axis off
    view([1 0 0])
    xlabel('x (um)'), ylabel('y (um)'), zlabel('z (um)')
    title(sprintf('row %d\navg ht: %.1f\navg rd: %.1f',i,ah(i)*rf,ar(i)*rf))
    set(gca,'Color','black')
end

% TGs = [Ts; Gs];
% mnts = min(TGs);
% mxts = max(TGs);
% mmTs = max([abs(mnts); abs(mxts)]);
% meanTGs = mean(TGs);
% Ts = Ts-repmat(meanTGs,[size(Ts,1) 1]);
% Ts = Ts/max(mmTs)*min([mmtsxy mmts(3)]);
% Gs = Gs-repmat(meanTGs,[size(Gs,1) 1]);
% Gs = Gs/max(mmTs)*min([mmtsxy mmts(3)]);

% subplot(1,5,5)
% plot3(Ts(:,1),Ts(:,2),Ts(:,3),'.','Color',0.5*[1 1 1],'MarkerSize',5), hold on
% plot3(Gs(:,1),Gs(:,2),Gs(:,3),'*','Color',[1 1 0],'MarkerSize',1), hold off
% axis([-mmtsxy mmtsxy -mmtsxy mmtsxy -mmts(3) mmts(3)])
% axis equal
% axis off

Link = linkprop(axs,{'CameraUpVector','CameraPosition','CameraTarget'});
setappdata(gcf,'StoreTheLink',Link)
rotate3d on
saveas(gcf,[outFiguresPath filesep sprintf('10_AggregatedTowersAndGoldPart.fig')]);
close all

%% save aggregated towers (convex hull) and gold particles to disk
disp('save aggregated towers (convex hull) and gold particles to disk')

summaryCell = cell(4,3);
cellIR = {ir1,ir2,ir3,ir4};
cellHR = {hr1,hr2,hr3,hr4};
cellRR = {rr1,rr2,rr3,rr4};
for row = 1:4
    AT = AggregatedTowers{row};
    AGP = AggregatedGoldParticles{row};
    if ~isempty(AT)
        k = convhull(AT(:,1),AT(:,2),AT(:,3));
        idx = unique(k(:));
        AT = AT(idx,:);

        TableAT = array2table(AT,'VariableNames',{'x','y','z'});
        writetable(TableAT,[outTablesPath filesep sprintf('AgrTwrRow%d.csv',row)]);

        TableHtRad = array2table([double(cellIR{row}') rf*cellHR{row}' rf*cellRR{row}'],'VariableNames',{'id','height','radius'});
        writetable(TableHtRad,[outTablesPath filesep sprintf('HtRadRow%d.csv',row)]);
    end
    if ~isempty(AGP)
        TableAGP = array2table(AGP,'VariableNames',{'x','y','z'});
        writetable(TableAGP,[outTablesPath filesep sprintf('AgrGPRow%d.csv',row)]);
    end
    summaryCell{row,1} = row;
    summaryCell{row,2} = ah(row)*rf;
    summaryCell{row,3} = ar(row)*rf;
end
T = cell2table(summaryCell,'VariableNames',{'row','avg_height','avg_radius'});
writetable(T,[outTablesPath filesep 'AvgHeightAndRadius.xls']);

%%
toc

%% [optional] visualize aggregated towers (convex hull) and gold particles using volumeViewer
% requires Matlab R2017A
% for best results, set VOLUME RENDERING option to MRI MIP and BACKGROUND COLOR to black
% works best on Windows

if 0
row = 4;

AT = AggregatedTowers{row};
AGP = AggregatedGoldParticles{row};
All = [AT; AGP];
MX = max(abs(All(:)));
AT = AT/MX*100;
AGP = AGP/MX*100;

k = convhull(AT(:,1),AT(:,2),AT(:,3));
idx = unique(k(:));
AT = AT(idx,:);

dt = delaunayTriangulation(AT);
mn = floor(min(AT))-1; mx = ceil(max(AT))+1;
[X,Y,Z] = meshgrid(mn(1):mx(1),mn(2):mx(2),mn(3):mx(3));
simplexIndex = pointLocation(dt,X(:),Y(:),Z(:));
mask = ~isnan(simplexIndex);
mask = 0.1*double(reshape(mask,size(X)));

AGP = ceil(AGP-repmat(mn,[size(AGP,1) 1])+1);

for i = 1:size(AGP,1)
    if AGP(i,2) > 0 && AGP (i,2) <= size(mask,1) && ...
       AGP(i,1) > 0 && AGP (i,1) <= size(mask,2) && ...
       AGP(i,3) > 0 && AGP (i,3) <= size(mask,3)
        mask(AGP(i,2),AGP(i,1),AGP(i,3)) = 1;
    end
end

volumeViewer(mask)
end